import numpy as np
from quakeio.utils.parseutils import open_quake


def read(read_file, **kwds):
    with open_quake(read_file,"r") as f:
        time_step = next(next(f))
        accel = np.genfromtxt(f, skip_header=1, delimiter=",", usecols=0)
    return GroundMotionRecord(
        GroundMotionSeries(accel), 
        meta={"accel.time_step": dt}
    )

def write(write_file, motion, /, label="eqsig data", **kwds):
    print(dir(motion.accel))
    print(dir(motion))
    time_step = motion.accel.time_step
    data = [
        label,
        f"{len(motion.accel)} {time_step}",
        *map(str,motion.accel)
    ]
    with open_quake(write_file, "w") as f:
        f.write("\n".join(data))

FILE_TYPES = {
    "eqsig": {
        "read": read,
        "write": write
    }
}




import re
from pathlib import Path


import numpy as np


class RealNumber(float):
    def __new__(cls, value, units=None):
        return float.__init__(cls, value)

    def __init__(self, value, units=None, time=None):
        self.units = units


class GroundMotionEvent(dict):
    """
    Container of GroundMotionRecord objects
    """

    def __init__(self, key, *args, event_date=None, **kwds):
        dict.__init__(self, **{record[key]: record for record in args})
        self.event_date = event_date

    def serialize(self) -> None:
        for record in self.values():
            record.serialize()


class GroundMotionRecord(dict):
    def __init__(self, accel, veloc, displ, meta={}):
        self.accel = accel
        self.displ = displ
        self.veloc = veloc
        dict.__init__(self, **meta)

    def serialize(self) -> None:
        self["accel"] = self.accel
        self["veloc"] = self.veloc
        self["displ"] = self.displ


class GroundMotionSeries(np.ndarray):
    def __new__(cls, input_array, metadata={}):
        obj = np.asarray(input_array).view(cls)
        for k,v in metadata.items():
            if not hasattr(obj,k):
                setattr(obj,k,v)
        return obj

    def __array_finalize__(self, obj):
        for k,v in self.__dict__.items():
            setattr(obj,k,v)

    def plot(self, ax=None, fig=None):
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(self)



# Claudio Perez
__version__ = "0.0.1"

from pathlib import Path

from . import csmip, nga, eqsig, basic_formats


FILE_TYPES = {}
FILE_TYPES.update(csmip.FILE_TYPES)
FILE_TYPES.update(eqsig.FILE_TYPES)
FILE_TYPES.update(basic_formats.FILE_TYPES)

DEFAULT_TYPES = {
    ".zip": "csmip.zip",
    ".v2": "csmip.v2",
    ".json": "json.record",
}


def read(read_file, *args, **kwds):
    """
    Generic ground motion reader
    """
    if "input_format" in kwds and kwds["input_format"]:
        typ = kwds["input_format"]
    else:
        try:
            typ = DEFAULT_TYPES[Path(read_file).suffix]
        except KeyError:
            raise ValueError("Unable to deduce input format")
    return FILE_TYPES[typ]["read"](read_file, *args, **kwds)


def write(write_file, ground_motion, write_format=None, *args, **kwds):
    if write_format:
        typ = write_format
    else:
        try:
            typ = DEFAULT_TYPES[Path(write_file).suffix]
        except KeyError:
            raise ValueError("Unable to deduce output format")
    FILE_TYPES[typ]["write"](write_file, ground_motion, *args, **kwds)





#import anon
import anabel
import anabel.backend as anp
try:
    from jax import jit
except:
    jit = lambda x: x

@anabel.template.template(dim="shape",order=0, main="step",origin="origin")
def linear_accel_scalar(df, dt, *,
    mass,
    damp,
    beta=1/4,
    gamma=1/2,
    tol = 1e-6,
    initial=(0.0, 0.0, 0.0)
):
    """Generate a 1D acceleration-driven Newmark integral.

    This function generates a Newmark integral from a 
    zero-order map, `f`, and two matrices `mass` and `damp`.
    It was adaped from accel_no1 to return only displacements.

    Parameters
    ----------

    Studies
    -------

    created 2021-06-14
    """
    stif = df
    fshape = [[0, 0]]
    shape = (1, (fshape[0], fshape[0],fshape[0]))
    state = {
        "u": 0.0,
        "v": 0.0,
        "a": 0.0,
    }

    origin = 0.0, 0, state


    C1 = 1/(beta*dt**2)*mass + gamma/(beta*dt)*damp
    C2 = 1/(beta*dt)*mass + (gamma/beta-1)*damp
    C3 = (1/(beta*2)-1)*mass + dt*(gamma/(2*beta)-1)*damp
    @jit
    def step(state, ds,  params={}, **kwds):
        u0, du0, ddu0 = state["u"], state["v"], state["a"]
        p_hat = mass*ds + C1*u0 + C2*du0 + C3*ddu0
        # Initial evaluation 
        fu0 = stif*u0
        res = p_hat - fu0 - C1*u0
        Du = res/(stif+C1) 
        u = u0 + Du

        du = gamma/(beta*dt)*(u-u0) + (1-gamma/beta)*du0 + dt*(1-gamma/(2*beta))*ddu0
        ddu = 1/(beta*dt**2)*(u-u0) -1/(beta*dt)*du0-(1/(2*beta)-1)*ddu0
        return {"u": u,"v": du, "a": ddu}, ddu
     
    return locals()



@anabel.template.template(dim="shape",order=0, main="step",origin="origin")
def linear_accel_vector(df, dt, *,
    mass,
    damp,
    beta=1/4,
    gamma=1/2,
    tol = 1e-6,
    initial=None
):
    """Generate a 1D acceleration-driven Newmark integral.

    This function generates a Newmark integral from a 
    zero-order map, `f`, and two matrices `mass` and `damp`.
    It was adaped from accel_no1 to return only displacements.

    Parameters
    ----------

    Studies
    -------

    created 2021-06-14
    """
    norm = anp.linalg.norm
    try:
        fshape = len(df)
    except TypeError:
        fshape = [[0, 0]]
    shape = (1, (fshape[0], fshape[0],fshape[0]))
    state = {
        "u": anp.zeros((1,1)),
        "v": anp.zeros((1,1)),
        "a": anp.zeros((1,1)),
    }
    if initial is None:
        initial = anp.zeros((3,*fshape[0]))

    origin = 0.0, anp.zeros((1,1)), state

    if isinstance(df,float):
        K_matrix = anp.array([[df]])
        mass = anp.array([[mass]]) 
        damp = anp.array([[damp]]) 

    C1 = 1/(beta*dt**2)*mass + gamma/(beta*dt)*damp
    C2 = 1/(beta*dt)*mass + (gamma/beta-1)*damp
    C3 = (1/(beta*2)-1)*mass + dt*(gamma/(2*beta)-1)*damp

    def step(ds, x0, state=state, params={}, **kwds):
        u0 = x0 
        u0, du0, ddu0 = state["u"], state["v"], state["a"]
        p_hat = mass*ds + C1@u0 + C2@du0 + C3@ddu0
        # Initial evaluation 
        fu0 = K_matrix@u0
        res = p_hat - fu0 - C1@u0
        Du = anp.linalg.solve(K_matrix+C1, res)
        u = u0 + Du

        du = gamma/(beta*dt)*(u-u0) + (1-gamma/beta)*du0 + dt*(1-gamma/(2*beta))*ddu0
        ddu = 1/(beta*dt**2)*(u-u0) -1/(beta*dt)*du0-(1/(2*beta)-1)*ddu0
        return u, {"u": u,"v": du, "a": ddu}
    
    return locals()





import io
from pathlib import Path
from typing import Union, IO
import contextlib


@contextlib.contextmanager
def open_quake(file, mode=None, archive=None):
    import sys
    if file == "-":
        if mode is None or mode == "" or "r" in mode:
            fh = sys.stdin
        else:
            fh = sys.stdout
    elif not archive:
        fh = open(file, mode)
    else:
        fh = archive.open(file, mode)

    # if file != "-":
    #     yield fh
    #     fh.close()
    try:
        yield fh
    finally:
        if file != "-":
            fh.close()


def parse_sequential_fields(data, field_spec: dict, parsed_fields={}) -> dict:
    """ """
    field_iterator = iter(field_spec.items())
    fields, (typs, regex) = next(field_iterator)
    for line in data:
        match = regex.findall(str(line))
        if match:
            prefix = ""
            assert len(fields) == len(typs) == len(match[0]), (regex, match[0])
            for field, typ, val in zip(fields, typs, match[0]):
                if field[0] == ".":
                    key = prefix + field
                else:
                    prefix = field
                    key = field

                parsed_fields[key] = typ(val)
            try:
                fields, (typs, regex) = next(field_iterator)
            except StopIteration:
                # All fields have been parsed out
                break
        else:
            pass
    return parsed_fields


def parse_sequential_fields_v0(data, fields, parsed_fields={}):
    """ """
    field_iterator = iter(fields.items())
    field, (keys, typs, regex) = next(field_iterator)

    for line in data:
        match = regex.findall(str(line))
        if match:
            parsed_fields[field] = {
                key: typ(val) for key, typ, val in zip(keys, typs, match[0])
            }
            try:
                field, (keys, typs, regex) = next(field_iterator)
            except StopIteration:
                # All fields have been parsed out
                break
    return parsed_fields


def get_file_type(
    file: Union[str, Path, IO], file_type: str, module: str = None
) -> str:
    if isinstance(file, io.IOBase):
        return file_type




import numpy as np
from scipy.integrate import solve_ivp


import quakeio
from quakeio.utils.processing import get_time_series, get_time_step
from .newmark import linear_accel_vector,linear_accel_scalar

class ResponseHistory:
    def __init__(self, accel, time_series=None, time_step=None):
        if isinstance(accel, (quakeio.core.GroundMotionSeries, np.ndarray)):
            self.accel = accel
        else:
            # a Ground motion record was supplied
            self.accel = accel.accel

        self.time_series = get_time_series(self.accel, time_series, time_step)
    
    def __call__(self, period, damping=1.0, R=1.0):
        #g = 386.4
        g = 9.80665*100.
        accel = self.accel
        omega = (2*np.pi)/period
        time = self.time_series
        t_span = (time[0], time[-1])
        u0 = (0.0, 0.0)

        def EOM(t, y):
            i = np.argmin(np.abs(time - t))
            D, dDdt = y
            dydt = [dDdt, -accel[i]*g-omega**2*D-2*damping*omega*dDdt]
            return dydt

        rtol = atol = 1e-8
        sol = solve_ivp(EOM, y0=u0, t_span=t_span, rtol=rtol, atol=atol)
        D = sol.y[0,:]
        t = sol.t
        return t, D

def response_spectrum(accel, periods=None, time_step=None, damping=1.0):
    import elle.transient
    import anabel, jax
    if periods is None:
        periods = np.linspace(0.1,5,100)
    mass = 1.0
    spect = []
    for period in periods: 
        k = (2.0*np.pi/period*np.sqrt(mass))**2
        integrator = linear_accel_scalar(
            df = k,
            dt = get_time_step(accel, time_step=time_step),
            mass = mass,
            damp = damping*2*np.sqrt(mass*k)
        )
        u_max = 0.0
        _, u, state = integrator.origin
        for a in accel:
            newstate, u = integrator(state, a)
            u_max = abs(u) if abs(u) > u_max else u_max
        spect.append(u_max)
    return np.array(spect).flatten()

 



# Claudio Perez
# 2021
"""
Parse a CSMIP Volume 2 strong motion data file.
"""
import re
import zipfile
from pathlib import Path
from typing import Union
from collections import defaultdict

import numpy as np

from .core import GroundMotionEvent, GroundMotionRecord, GroundMotionSeries, RealNumber
from .utils.parseutils import parse_sequential_fields, get_file_type, open_quake

NUM_COLUMNS = 8
HEADER_END_LINE = 45

# Regular expression for extracting decimal number
RE_DECIMAL = "[-]?[0-9]*[.][0-9]*"
# Regular expression for extracting units
RE_UNITS = "[a-z,/,*,0-9]*"

date = lambda x: x

# fmt: off
HEADER_FIELDS = {
    #("record.date_processed",): ((str,),
    #   re.compile(r"Processed: *(\d\d/\d\d/\d\d,)"),
    #),
    ("record.station_no", "record.azimuth"): ((str,str),
        re.compile(
            rf"Station No\. *([0-9]*) *({RE_DECIMAL}[NSEW]*, *{RE_DECIMAL}[NSEW]*)"
        )
    ),
    ("record.instr_period", ".units"): ((float,str),
        re.compile(
            rf"Instr Period = ({RE_DECIMAL}) ({RE_UNITS}),"
        )
    ),
    ("record.peak_accel", ".units", ".time"): ((float, str, float),
        re.compile(
            rf"Peak *acceleration *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})"
        )
    ),
    ("record.peak_veloc",".units",".time"): ((float, str, float), 
        re.compile(
            rf"Peak *velocity *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})"
        )
    ),
    ("record.peak_displ",".units",".time"): ((float, str, float),
        re.compile(
            rf"Peak *displacement *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})"
        ),
    ),
    ("record.init_veloc",".units"): ((float, str),
        re.compile(rf"Initial velocity *= *({RE_DECIMAL}) *({RE_UNITS});"),
    ),
    ("accel.shape", "accel.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of accel data equally spaced at *({RE_DECIMAL})"),
    ),
    ("veloc.shape", "accel.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of veloc data equally spaced at *({RE_DECIMAL})"),
    ),
    ("displ.shape", "accel.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of displ data equally spaced at *({RE_DECIMAL})"),
    ),
}
# fmt: on


def read_event(read_file, **kwds):
    """
    Take the name of a CSMIP zip file and extract record data for the event.
    """
    zippath = Path(read_file)
    archive = zipfile.ZipFile(zippath)
    records = []
    for file in archive.namelist():
        if file.endswith(".v2"):
            records.append(read_record_v2(file, archive))
    metadata = {}
    return GroundMotionEvent("file_name", *records, **metadata)


def read_record_v2(
    read_file, archive: zipfile.ZipFile = None, summarize=False, **kwds
) -> GroundMotionRecord:
    """
    Read a ground motion record using the CSMIP Volume 2 format
    """
    filename = Path(read_file)
    # Parse header fields
    with open_quake(read_file, "r", archive) as f:
        header_data = parse_sequential_fields(f, HEADER_FIELDS)
    parse_options = dict(
        delimiter=10,  # fields are 10 chars wide
    )
    # Reopen and parse out data; Note, only the first call
    # provides a skip_header argument as successive reads
    # pick up where the previous left off.
    if not summarize:
        with open_quake(read_file, "r", archive) as f:
            accel = np.genfromtxt(
                f,
                skip_header=HEADER_END_LINE+1,
                max_rows=header_data["accel.shape"] // NUM_COLUMNS,
                **parse_options,
            ).flatten()
            next(f)
            veloc = np.genfromtxt(
                f, max_rows=header_data["veloc.shape"] // NUM_COLUMNS, **parse_options
            ).flatten()
            next(f)
            displ = np.genfromtxt(
                f, max_rows=header_data["displ.shape"] // NUM_COLUMNS, **parse_options
            ).flatten()
    else:
        accel, veloc, displ = None, None, None

    # Separate out metadata
    record_data = {}
    series_data = defaultdict(dict)
    for key, val in header_data.items():
        typ, k = key.split(".", 1)
        if typ == "record":
            record_data.update({k: val})
        elif typ in ["accel","veloc","displ"]:
            series_data[typ].update({k: val})

    record_data["file_name"] = filename.name
    return GroundMotionRecord(
        GroundMotionSeries(accel,series_data["accel"]),
        GroundMotionSeries(veloc,series_data["veloc"]),
        GroundMotionSeries(displ,series_data["displ"]),
        meta=record_data,
    )


FILE_TYPES = {
    "csmip.v1": {"type": GroundMotionRecord, "read": read_record_v2},
    "csmip.v2": {"type": GroundMotionRecord, "read": read_record_v2},
    "csmip.v3": {"type": GroundMotionRecord, "read": read_record_v2, "spec": ""},
    "csmip.zip": {"type": GroundMotionEvent, "read": read_event},
}


def read_record(read_file, *args):
    file = Path(read_file)
    if file.suffix == ".v2":
        return read_record_v2(file)


def read(read_file, input_format=None):
    # file_type = get_file_type(file,file_type,"csmip")
    # if isinstance(
    pass



import re

import numpy as np

RE_TIME_STEP = re.compile(r"DT=\s+(.*)SEC")


def parse_nga(file_name, meta=False):
    with open(file_name, "r") as f:
        header = [next(f) for _ in range(4)]
    dt = float(RE_TIME_STEP.search(header[-1]).group(1).strip())
    accel = np.genfromtext(file_name, skip_header=4, skip_footer=1).flatten()



import argparse
import quakeio


def build_parser():
    # fmt: off
    parser = argparse.ArgumentParser(
        prog="quakeio", description="Ground motion processing utilities."
    )
    parser.add_argument("read_file", nargs="?", default="-")
    parser.add_argument("-o", "--write_file", default="-")
    parser.add_argument("-c", "--command", action="append")
    parser.add_argument("-S", "--summarize", action="store_true")
    parser.add_argument(
        "-f",
        dest="input_format",
        help="Specity input file format",
        choices=list(quakeio.FILE_TYPES.keys()),
    )
    parser.add_argument(
        "-t",
        "--to",
        dest="write_format",
        help="Specity output file format",
        default="json.record",
    )
    parser.add_argument("-s", "--scale", help="Scale ground motions by factor FACTOR")

    parser.add_argument(
        "--version", help="Print version and exit.", action="store_true"
    )
    # fmt: on
    return parser


def cli(*args, **kwds):
    if kwds["version"]:
        print(quakeio.__version__)
        return
    if "command" in kwds:
        cmds = kwds.pop("command")
    # if kwds["read_file"] == "-":
    #    kwds["read_file"] = sys.stdin
    # write_file = sys.stdout if kwds["write_file"] == "-" else kwds["write_file"]
    write_file = kwds["write_file"]
    del kwds["write_file"]
    motion = quakeio.read(**kwds)
    quakeio.write(write_file, motion, **kwds)
    print("\n")


def main():
    args = build_parser().parse_args()
    cli(**vars(args))


if __name__ == "__main__":
    main()



import json

from quakeio.utils.parseutils import open_quake


class QuakeEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "tolist"):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def read_json(read_file, **k):

    return json.load(read_file)


def write_json(write_file, ground_motion, indent=4, summarize=True, **kwds):
    if not summarize:
        ground_motion.serialize()
    with open_quake(write_file, "w") as f:
        json.dump(ground_motion, f, indent=indent, cls=QuakeEncoder)


def read_yaml(read_file, **k):
    import yaml

    return yaml.load(read_file, loader=yaml.Loader)


def write_yaml(write_file, ground_motion, summarize=True, **kwds):
    import yaml

    if not summarize:
        ground_motion.serialize()

    with open_quake(write_file, "w") as f:
        yaml.dump(ground_motion, f)


FILE_TYPES = {
    "json": {"write": write_json},
    "yaml": {"write": write_yaml},
    "json.record": {"read": read_json, "write": write_json},
    "yaml.record": {"read": read_yaml, "write": write_yaml},
}
